(*
	process OCR on a pdf file;
	requires:
		- tesseract [depending on imagemagick and ghostscript]
		- unpaper
		- hocr2pdf (from ExactImage)
	
	(C) 2010-2013 Tobias Elze, modified for tesseract Heinrich Schwietering 2012
	patch for -rgb option contributed by James Cort
*)

include Pdfsandwich_version;;		(*provides string variable pdfsandwich_version*)

(*default binary names:*)
let unpaper = ref "unpaper";;
let convert = ref "convert";;
let tesseract = ref "tesseract";;
let gs = ref "gs";;
let hocr2pdf = ref "hocr2pdf";;

(*global flags:*)
let verbose = ref false;;
let quiet = ref false;;

(*print output, if verbose option is set (default)*)
let pr s =
	if not !quiet then print_endline s
;;

(*execute command cmd and print it's invocation line (if verbose is set):*)
let run cmd =
	if !verbose then pr cmd;
	ignore (Sys.command cmd);
;;

(*check if binary bin exists (in search path):*)
let check_for_binary bin =
	try
	let in_ch = Unix.open_process_in ("which " ^ bin) in
	let s = input_line in_ch in
	if s = "" then failwith "";
	ignore (Unix.close_process_in in_ch);
	with _ -> failwith 
		("Could not find program " 
		^ bin 
		^ ". Make sure this program exists and can be found in your search path.\nUse command line options to specify a custom binary.")
;;

(*return number of pages of a PDF file; needs gs:*)
let number_of_pages filename =
	try
	let in_ch = Unix.open_process_in
		("printf \"/PDFfile PDFname (r) file def\n\n/PageCountString 255 string\n\ndef\nsystemdict /.setsafe known { .setsafe } if\n/.show.stdout { (%%stdout) (w) file } bind def\n/puts { .show.stdout exch writestring } bind def\nGS_PDF_ProcSet begin\npdfdict begin\nPDFfile\npdfopen begin\n/FirstPage where { pop } { /FirstPage 1 def } ifelse\n/LastPage where { pop } { /LastPage pdfpagecount def } ifelse\n() puts\nLastPage FirstPage sub 1 add\n\nPageCountString cvs puts\n(\\\\\\\\n) print\nquit \n\""
		^ "|" ^ !gs ^ " -q -sNODISPLAY -sPDFname=\""
		^ filename ^ "\" -")
	in
	let pagestring = input_line in_ch in
	ignore (Unix.close_process_in in_ch);
	int_of_string pagestring;
	with _ -> failwith ("Error: Could not determine number of pages of file " ^ filename)
;;

(*return number of CPUs:*)
let number_of_cpus () =
	(*different code for different platforms;
	runtime OS detection:*)
	try
	(match Sys.os_type with
	| "Unix" ->
	(
		let scriptstring = 
			let p = Unix.open_process_in "uname" in
			let os = input_line p in
			ignore (Unix.close_process_in p);
			match os with 
			| "Darwin" -> "sysctl -n hw.ncpu"
			| _ -> "cat /proc/cpuinfo | grep processor | awk '{a++} END {print a}'" (*Linux*)
		in
		let in_ch = Unix.open_process_in scriptstring in
		let numstr = input_line in_ch in
		ignore (Unix.close_process_in in_ch);
		int_of_string numstr;
	)
	| _ -> failwith "Not yet implemented for Non-Unix systems.")
	with _ -> 1
;;

(*process OCR on pdf file infile and save the results to outfile:*)
let process_ocr infile outfile first_page last_page resolution rgb nthreads language convertopts tessopts hocropts preprocess unpaperopts debug enforcehocr2pdf =
	let pages_to_process = last_page - first_page + 1 in
	let hocr_resolution = Str.global_replace (Str.regexp "^\\(.+\\)x.*$") "\\1" resolution in
	if nthreads > 1 then
		pr ("\nParallel processing with " ^ (string_of_int nthreads) ^ " threads started.\nProcessing page order may differ from original page order.\n");
	
	let process_page (curr_page, pdfname) =
		let tmppicfile = Filename.temp_file "pdfsandwich" (if rgb then ".ppm" else ".pbm") in
		let tmpocrfile = Filename.temp_file "pdfsandwich" "" in
		let tmpcolfigfile = Filename.temp_file "pdfsandwich" "_col.png" in
		let tmpunpaperfile = Filename.temp_file "pdfsandwich" ("_unpaper" ^ (if rgb then ".ppm" else ".pbm")) in
		if not !quiet then
			Printf.printf "Processing page %i.\n" curr_page;
		flush_all ();
		let convoptstmp =
			if rgb then " -depth 8 -background white -flatten +matte -density " else " -type Bilevel  -density "
		in
		run (!convert ^ " " ^ convertopts ^ convoptstmp ^ resolution ^ " \"" ^ infile ^ "[" ^ (string_of_int (curr_page-1)) ^ "]\" " ^ tmppicfile);
		let tessout = if not !verbose then ">/dev/null 2>&1" else "" in
		let tessinputfile = 
			if preprocess then
			(
				run (!unpaper ^ " --overwrite " ^ unpaperopts ^ " " ^ tmppicfile ^ " " ^ tmpunpaperfile ^ tessout);
				tmpunpaperfile
			)
			else tmppicfile
		in
		
		(*test if tesseract can output pdf files:*)
		run (!tesseract ^ " " ^ tessinputfile ^ tessout ^ " " ^ tmpocrfile ^ " " ^ tessopts ^ " -l " ^ language ^ " pdf ");
		
		if (not enforcehocr2pdf) && Sys.file_exists (tmpocrfile ^ ".pdf") then
		(
			if !verbose then pr ("Renaming " ^ (tmpocrfile ^ ".pdf") ^ " to " ^ pdfname);
			Sys.rename (tmpocrfile ^ ".pdf") pdfname;
		)
		else
		(
			if not !quiet then
				print_endline "Version of tesseract is prior to 3.03 and cannot output pdf yet. Using hocr2pdf instead.";
			run (!tesseract ^ " " ^ tessinputfile ^ tessout ^ " " ^ tmpocrfile ^ " " ^ tessopts ^ " -l " ^ language ^ " hocr ");
			let hocrinputfile = if Sys.file_exists (tmpocrfile ^ ".html") then tmpocrfile ^ ".html" else tmpocrfile ^ ".hocr" in
			run (!hocr2pdf ^ hocropts ^ " -r " ^ hocr_resolution ^ " -i " ^ tessinputfile ^ " -o " ^ pdfname ^ "<" ^ hocrinputfile);
			if not debug then Sys.remove hocrinputfile;
		);
		
		let rm_if_exists f = if Sys.file_exists f then Sys.remove f in
		if not debug then
		(
			rm_if_exists (tmpocrfile ^ ".pdf");
			rm_if_exists tmppicfile;
			rm_if_exists tmpocrfile;
			rm_if_exists tmpcolfigfile;
			rm_if_exists tmpunpaperfile;
		);
	in
	let process_pagelist = Array.iter process_page in
	
	let tmppdf_arr = 
		Array.init pages_to_process (fun i -> (i+first_page, Filename.temp_file "pdfsandwich" ".pdf"))
	in
	let intdiv = pages_to_process / nthreads
	and remainder = pages_to_process mod nthreads in
	
	let nested_arr =
		let rec f l ncum rmndr = function
			| 0 -> Array.of_list l
			| i -> 
				let npgs = intdiv + (if rmndr>0 then 1 else 0) in
				let ncum1 = ncum + npgs in
				let new_el = Array.sub tmppdf_arr ncum npgs in
				f (new_el::l) ncum1 (rmndr-1) (i-1)
		in
		f [] 0 remainder nthreads
	in

	let threadarr = Array.map (Thread.create process_pagelist) nested_arr in
	Array.iter Thread.join threadarr;
		
	let pdffilenamelist = List.map snd (Array.to_list tmppdf_arr) in
	let pdfliststring = String.concat " " pdffilenamelist in
	pr ("OCR done. Writing \"" ^ outfile ^ "\"");
	run (!gs ^ " -q -dNOPAUSE -dBATCH -sDEVICE=pdfwrite -sOutputFile=\"" ^ outfile ^ "\" " ^ pdfliststring);
	if (not debug) then List.iter Sys.remove pdffilenamelist;
;;

let main () =
	let arg_filename = ref "" in
	let outputfile = ref "" in
	let first_page = ref 1
	and last_page = ref 0 in
	let resolution = ref "300x300" in
	let lang = ref "eng" in
	let rgb = ref false in
	let preprocess = ref true in
	let unpaperopts = ref "" in
	let hocropts = ref "" in
	let convertopts = ref "" in	(*additional convert options*)
	let tessopts = ref "" in	(*additional tesseract options*)
	let nthreads = ref 0 in
	let debug = ref false in
	let enforcehocr2pdf = ref false in
	let set_hocr_opts op () = hocropts := !hocropts ^ " " ^ op in
		
	let speclist = [
		("-convert", Arg.Set_string convert, "\t -convert filename : name of convert binary (default: convert)");
		("-coo", Arg.Set_string convertopts, "\t\t -coo options : additional convert options; make sure to quote;\n\t\t  e.g. -coo \"-normalize -black-threshold 75%\"\n\t\t  call convert --help or man convert for all convert options");
		("-debug", Arg.Set debug, "\t keep all temporary files in /tmp (for debugging)");
		("-enforcehocr2pdf", Arg.Set enforcehocr2pdf, "\t use hocr2pdf even if tesseract >= 3.03");
		("-first_page", Arg.Set_int first_page, "\t -first_page number : number of page to start OCR from (default: 1)");
		("-gs", Arg.Set_string gs, "\t\t -gs filename : name of gs binary (default: gs)");
		("-hocr2pdf", Arg.Set_string hocr2pdf, "\t -hocr2pdf filename : name of hocr2pdf binary (default: hocr2pdf);\n\t\t  ignored for tesseract >= 3.03 unless option -enforcehocr2pdf is set");
		("-hoo", Arg.String (fun s -> set_hocr_opts s()), "\t\t -hoo options : additional hocr2pdf options; make sure to quote");
		("-last_page", Arg.Set_int last_page, "\t -last_page number : number of page up to which to process OCR (default: number of pages in inputfile)");
		("-lang", Arg.Set_string lang, "\t -lang language : language of the text; option to tesseract (defaut: eng)\n\t\t  e.g: eng, deu, deu-frak, fra, rus, swe, spa, ita...\n\t\t  see http://code.google.com/p/tesseract-ocr/downloads/list\n\t\t  or LANGUAGE section of man page;\n\t\t  Multiple languages may be specified, separated by plus characters.");
		("-noimage", Arg.Unit (set_hocr_opts "-n"), "\t do not place the image over the text (requires hocr2pdf; ignored without -enforcehocr2pdf option)");
		("-nopreproc", Arg.Clear preprocess, "\t do not preprocess with unpaper");
		("-nthreads", Arg.Set_int nthreads, "\t -nthreads number : number of parallel threads (default: guessed number of CPUs; if guessing fails: 1)");
		("-o", Arg.Set_string outputfile, "\t\t -o filename : output file; default: inputfile_ocr.pdf");
		("-resolution", Arg.Set_string resolution, "\t -resolution NUMxNUM : resolution used for OCR (default: 300x300)");
		("-rgb", Arg.Set rgb, "\t\t use RGB color space for images (default: black and white);\n\t\t  use with care: causes problems with some color spaces");
		("-sloppy_text", Arg.Unit (set_hocr_opts "-s"), "\t sloppily place text, group words, do not draw single glyphs;\n\t\t  ignored for tesseract >= 3.03 unless option -enforcehocr2pdf is set");
		("-tesseract", Arg.Set_string tesseract, "\t -tesseract filename : name of tesseract binary (default: tesseract)");
		("-tesso", Arg.Set_string tessopts, "\t -tesso options : additional tesseract options; make sure to quote");
		("-unpaper", Arg.Set_string unpaper, "\t -unpaper filename : name of unpaper binary (default: unpaper)");
		("-unpo", Arg.Set_string unpaperopts, "\t -unpo options : additional unpaper options; make sure to quote");
		("-quiet", Arg.Set quiet, "\t suppress output");
		("-verbose", Arg.Set verbose, "\t produce more output");
		("-version", Arg.Unit (fun () -> Printf.printf "pdfsandwich version %s\n" pdfsandwich_version; exit 0), "\t print version and quit");
	] in
	
	Arg.parse speclist (fun s -> arg_filename := s) "USAGE: pdfsandwich [options] inputfile.pdf\n\nOptions:";
	List.iter check_for_binary [!convert; !tesseract; !gs; !hocr2pdf];
	if !outputfile = "" then outputfile := Str.global_replace (Str.regexp "^\\(.+\\)\\.pdf$") "\\1_ocr.pdf" !arg_filename;
	(try Unix.access !arg_filename [Unix.R_OK] with _ -> failwith ("Could not open file " ^ !arg_filename)); 
	pr ("pdfsandwich version " ^ pdfsandwich_version);
	pr ("Input file: \"" ^ !arg_filename ^ "\"");
	pr ("Output file: \"" ^ !outputfile ^ "\"");
	let npages = number_of_pages !arg_filename in
	pr ("Number of pages in inputfile: " ^ (string_of_int npages));
	if (!first_page < 1 || !first_page > npages) then failwith ("Value " ^ (string_of_int !first_page) ^ " is invalid as first_page.");
	if !last_page < 1 then last_page:= npages;
	if (!last_page < !first_page || !last_page > npages) then failwith ("Value " ^ (string_of_int !last_page) ^ " is invalid as last_page.");
	if !nthreads < 1 then (*guess number of CPUs*)
		nthreads := number_of_cpus ();
	let npages_to_process = !last_page - !first_page + 1 in
	if !nthreads > npages_to_process then
	(
		pr ("More threads than pages. Using " ^ (string_of_int npages_to_process) ^ " threads instead.");
		nthreads := npages_to_process;
	);
	
	process_ocr !arg_filename !outputfile !first_page !last_page !resolution !rgb !nthreads !lang !convertopts !tessopts !hocropts !preprocess !unpaperopts !debug !enforcehocr2pdf;
	pr "\nDone.";
;;

main ();;


